import bblean

# General variables used in |substitutions|
project = "BitBIRCH-Lean"
copyright = "2025  The Miranda-Quintana Lab and other BitBirch developers"
author = "The Miranda-Quintana Lab and other BitBirch developers"
_version = ".".join(bblean.__version__.split(".")[:2])
version = f"{_version} (dev)" if "dev" in bblean.__version__ else _version
release = bblean.__version__

# Common substitutions used throughout the docs
rst_epilog = """
..  |X| replace:: Fingerprints
"""
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",  # For autogen python module docs
    "sphinx.ext.intersphinx",
    "sphinx.ext.githubpages",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",  # For google-style docstr
    "sphinx_design",  # For grid directive
    "nbsphinx",
]
# Autosummary
templates_path = ["_templates"]
autosummary_ignore_module_all = False  # Respect <module>.__all__
# Extensions config
# autodoc
autodoc_typehints_format = "short"  # Avoid qualified names in autodoc types
autodoc_typehints = "description"  # Write types in description, not in signature
autodoc_typehints_description_target = "documented"  # Only write type for docum. params
autodoc_inherit_docstrings = True  # Docstring of supercls is used by dflt
autodoc_default_options = {
    "members": None,  # This means "True"
    "member-order": "bysource",  # Document in the same order as python source code
}
# napoleon
napoleon_google_docstring = True
napoleon_numpy_docstring = True
# intersphinx
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "rdkit": ("https://www.rdkit.org/docs/", None),
}


# General sphinx config
# nitpicky = True  # Fail if refs can't be resolved TODO re-enable and fix invalid refs
default_role = "any"  # Behavior of `inline-backticks`, try to link to "anything"
pygments_style = "sphinx"  # Code render style
master_doc = "index"  # Default, Main toctree
source_suffix = {".rst": "restructuredtext"}  # Default, Suffix of files

# Python-domain sphinx config
python_use_unqualifierd_type_names = True  # Try to dequa py obj names if resolveable
python_display_short_literal_types = True  # show literals as a | b | ...

# HTML config
html_title = f"{project} v{version} Manual"
html_static_path = ["_static"]  # Static html resources
html_css_files = ["style.css"]  # Overrides for theme style sheet
html_theme = "pydata_sphinx_theme"
html_use_modindex = True
html_domain_indices = False
html_copy_source = False
html_file_suffix = ".html"
htmlhelp_basename = "bblean-docs"

# PyData Theme config
# Primary HTML sidebar (left)
html_sidebars = {
    "index": [],
    "installing": [],
    "user-guide": ["sidebar-nav-bs"],
    "linux_memory_setup": [],
    "api-reference": ["sidebar-nav-bs"],
    "api_autogen/*": ["sidebar-nav-bs"],
    "publications": [],
}
html_theme_options = {
    "show_toc_level": 1,  # default is 2?
    "primary_sidebar_end": [],
    "navbar_center": ["navbar-nav"],
    # Secondary HTML sidebar (right)
    "secondary_sidebar_items": {
        "index": [],
        "installing": [],
        "user-guide": [],
        "linux_memory_setup": [],
        "api-reference": ["page-toc"],
        "_api_autogen/*": ["page-toc"],
        "publications": [],
    },
    # Misc
    "github_url": "https://github.com/mqcomplab/bblean",
    "icon_links": [],
    "logo": {
        "image_light": "_static/logo-light-bw.svg",
        "image_dark": "_static/logo-dark-bw.svg",
    },
    "show_version_warning_banner": True,
}

# Other: info, tex, man
latex_documents = [
    (
        master_doc,
        "BitBIRCH-Lean.tex",
        "BitBIRCH-Lean Documentation",
        "BitBIRCH-Lean developers",
        "manual",
    ),
]
man_pages = [(master_doc, "bblean", "BitBIRCH-Lean Documentation", [author], 1)]
texinfo_documents = [
    (
        master_doc,
        "BitBIRCH-Lean",
        "BitBIRCH-Lean Documentation",
        author,
        "BitBIRCH-Lean",
        "Clustering of huge molecular libraries",
        "Miscellaneous",
    ),
]



import os
import platform
from pathlib import Path
import tomli

import setuptools

root = Path(__file__).parent
with open(root / "pyproject.toml", "rb") as f:
    pyproject = tomli.load(f)

name = pyproject["project"]["name"]

_setup_kwargs = dict(
    name=name,
    use_scm_version=True,
    build_requires=["setuptools_scm"],
    long_description_content_type="text/markdown",
)

X86 = platform.machine().lower() in ["x86_64", "amd64", "i386", "x86", "i686"]
APPLE_SILICON = platform.machine().lower() in ["arm64", "aarch64"]
ARM = platform.machine().lower().startswith("arm") and not APPLE_SILICON

# Build C++ extensions (recommended)
extra_compile_args = []
if os.getenv("BITBIRCH_BUILD_CPP"):
    import pybind11
    from pybind11.setup_helpers import Pybind11Extension, WIN

    # setuptools paths must be relative
    ext_sources = [str((Path(name) / "csrc" / "similarity.cpp"))]
    if not WIN:
        extra_compile_args.append("-O3")  # -O3 includes -ftree-vectorize
    if not WIN:
        if X86:
            if os.getenv("BITBIRCH_BUILD_NATIVE"):
                extra_compile_args.extend(
                    ["-march=native", "-mtune=native", "-mpopcnt"]
                )
            else:
                extra_compile_args.extend(
                    ["-march=nocona", "-mtune=haswell", "-mpopcnt"]
                )
        elif APPLE_SILICON:
            if os.getenv("BITBIRCH_BUILD_NATIVE"):
                extra_compile_args.extend(["-mcpu=native"])
    if os.getenv("BITBIRCH_BUILD_CUSTOM_FLAGS"):
        # Override defaults and allow user to specify all flags, useful for attempting
        # to compile with MSVC
        extra_compile_args = []

    # NOTE: Compile with -DDEBUG_LOGS=1 to get some debug info from the extensions
    if os.getenv("BITBIRCH_DEBUG_EXT"):
        extra_compile_args.append("-fopt-info-vec-all")  # print loop vectorization info
        extra_compile_args.append("-DDEBUG_LOGS=1")
    _setup_kwargs["ext_modules"] = [
        Pybind11Extension(
            ".".join((name, "_cpp_similarity")),
            ext_sources,
            include_dirs=[pybind11.get_include()],
            language="c++",
            cxx_std=17,
            extra_compile_args=extra_compile_args,
        ),
    ]

setuptools.setup(**_setup_kwargs)



import os
import numpy as np
import pytest

from inline_snapshot import snapshot

# TODO: Fix the tests with pytest-subtests so that both the _py_similarity and the
# _cpp_similarity are tested independently
import bblean.similarity
import bblean._py_similarity as pysim

try:
    import bblean._cpp_similarity as csim

    CSIM_AVAIL = True
except ImportError:
    if os.getenv("BITBIRCH_CANT_SKIP_CPP_TESTS"):
        raise
    CSIM_AVAIL = False
from bblean.fingerprints import make_fake_fingerprints, unpack_fingerprints
from bblean.similarity import centroid_from_sum, centroid as centroid_from_fps
import bblean.similarity as gensim


def test_jt_most_dissimilar_packed() -> None:
    fps = make_fake_fingerprints(10, seed=17408390758220920002)
    expect_idx1 = 1
    expect_idx2 = 2
    expect_sims1 = np.array(
        [
            0.05083333,
            1.0,
            0.03805175,
            0.05077805,
            0.04651163,
            0.04683841,
            0.05954198,
            0.06254826,
            0.05578947,
            0.05006954,
        ]
    )
    expect_sims2 = np.array(
        [
            0.23452294,
            0.03805175,
            1.0,
            0.2352518,
            0.08961039,
            0.1166033,
            0.22281879,
            0.2363388,
            0.2045264,
            0.17490119,
        ]
    )
    (
        idx1,
        idx2,
        sims1,
        sims2,
    ) = pysim.jt_most_dissimilar_packed(fps)
    assert idx1 == expect_idx1
    assert idx2 == expect_idx2
    assert np.isclose(sims1, expect_sims1).all()
    assert np.isclose(sims2, expect_sims2).all()
    if CSIM_AVAIL:
        (
            idx1,
            idx2,
            sims1,
            sims2,
        ) = csim.jt_most_dissimilar_packed(fps)
        assert idx1 == expect_idx1
        assert idx2 == expect_idx2
        assert np.isclose(sims1, expect_sims1).all()
        assert np.isclose(sims2, expect_sims2).all()


@pytest.mark.skipif(not CSIM_AVAIL, reason="Requires C++ extensions")
def test_popcount_1d() -> None:
    fps_1d = make_fake_fingerprints(1, seed=17408390758220920002).reshape(-1)
    expect_1d = 1137
    out = csim._popcount_1d(fps_1d)
    assert (out == pysim._popcount(fps_1d)).all()
    assert out == expect_1d


@pytest.mark.skipif(not CSIM_AVAIL, reason="Requires C++ extensions")
def test_popcount_2d() -> None:
    fps_2d = make_fake_fingerprints(10, seed=17408390758220920002)
    expect_2d = [1137, 124, 558, 1159, 281, 323, 1264, 1252, 879, 631]
    out = csim._popcount_2d(fps_2d)
    assert (out == pysim._popcount(fps_2d)).all()
    assert out.tolist() == expect_2d


@pytest.mark.skipif(not CSIM_AVAIL, reason="Requires C++ extensions")
def test_cpp_centroid() -> None:
    fps = make_fake_fingerprints(10, seed=17408390758220920002, pack=False)
    _sum = fps.sum(0)
    num = len(fps)
    centroid = csim.centroid_from_sum(_sum, num, pack=False)
    expect_centroid = centroid_from_sum(_sum, num, pack=False)
    assert (centroid == expect_centroid).all()
    centroid = csim.centroid_from_sum(_sum, num, pack=True)
    expect_centroid = centroid_from_sum(_sum, num, pack=True)
    assert (centroid == expect_centroid).all()
    centroid = centroid_from_fps(fps, input_is_packed=False)
    assert (centroid == expect_centroid).all()


# TODO: Move this test
@pytest.mark.skipif(not CSIM_AVAIL, reason="Requires C++ extensions")
def test_cpp_unpacking() -> None:
    for seed in [
        17493821988544178123,
        4478748046060904849,
        4727712347772598054,
        15490537310413187550,
    ]:
        fps = make_fake_fingerprints(
            10, seed=17408390758220920002, pack=True, n_features=2024, dtype=np.uint8
        )
        expect_unpacked = unpack_fingerprints(fps)
        unpacked = csim._nochecks_unpack_fingerprints_2d(fps)
        assert (expect_unpacked == unpacked).all()
        unpacked = csim.unpack_fingerprints(fps)
        assert (expect_unpacked == unpacked).all()
        for fp in fps:
            expect_unpacked = unpack_fingerprints(fp)  # type: ignore
            unpacked = csim._nochecks_unpack_fingerprints_1d(fp)  # type: ignore
            assert (expect_unpacked == unpacked).all()
            unpacked = csim.unpack_fingerprints(fp)  # type: ignore
            assert (expect_unpacked == unpacked).all()


def test_jt_sim_packed() -> None:
    fps = make_fake_fingerprints(10, seed=17408390758220920002)
    first = fps[0]
    expect = np.array(
        [
            1.0,
            0.050833333333333,
            0.234522942461763,
            0.400854179377669,
            0.128980891719745,
            0.130030959752322,
            0.411522633744856,
            0.411104548139398,
            0.309090909090909,
            0.246826516220028,
        ],
        dtype=np.float64,
    )
    out = pysim._jt_sim_arr_vec_packed(fps, first)
    assert np.isclose(out, expect).all()

    if CSIM_AVAIL:
        out = csim._jt_sim_arr_vec_packed(fps, first)
        assert np.isclose(out, expect).all()

    # General
    out = gensim.jt_sim_packed(fps, first)
    assert np.isclose(out, expect).all()

    out = gensim.jt_sim_packed(first, fps)
    assert np.isclose(out, expect).all()

    out = gensim.jt_sim_packed(fps[0], first)
    assert out == expect[0]


def test_jt_isim_from_sum() -> None:
    fps = make_fake_fingerprints(100, seed=17408390758220920002, pack=False)
    c_total = fps.sum(0)
    c_objects = len(fps)
    s = pysim.jt_isim_from_sum(c_total, c_objects)
    assert s == 0.21824334501491158
    if CSIM_AVAIL:
        s = csim.jt_isim_from_sum(c_total, c_objects)
        assert s == 0.21824334501491158


def test_jt_isim() -> None:
    fps = make_fake_fingerprints(100, seed=17408390758220920002, pack=False)
    fps_packed = make_fake_fingerprints(100, seed=17408390758220920002, pack=True)
    s = pysim.jt_isim_unpacked(fps)
    assert s == 0.21824334501491158
    s2 = pysim.jt_isim_packed(fps_packed)
    assert s2 == 0.21824334501491158
    if CSIM_AVAIL:
        from bblean.similarity import (
            jt_isim_unpacked as jt_isim_unpacked_wrap,
            jt_isim_packed as jt_isim_packed_wrap,
        )

        s = csim.jt_isim_unpacked_u8(fps)
        assert s == 0.21824334501491158
        s2 = csim.jt_isim_packed_u8(fps_packed)
        assert s2 == 0.21824334501491158
        s = jt_isim_unpacked_wrap(fps)
        assert s == 0.21824334501491158
        s2 = jt_isim_packed_wrap(fps_packed)
        assert s2 == 0.21824334501491158


def test_jt_isim_from_sum_disjoint() -> None:
    fps = make_fake_fingerprints(1, seed=17408390758220920002, pack=False)
    disjoint = (~fps.astype(np.bool)).view(np.uint8)
    fps = np.concatenate((fps, disjoint))
    c_total = fps.sum(0)
    c_objects = len(fps)
    assert pysim.jt_isim_from_sum(c_total, c_objects) == 0
    if CSIM_AVAIL:
        assert csim.jt_isim_from_sum(c_total, c_objects) == 0

    fps = np.eye(2048, 2048, dtype=np.uint8)
    c_total = fps.sum(0)
    c_objects = len(fps)
    assert pysim.jt_isim_from_sum(c_total, c_objects) == 0
    if CSIM_AVAIL:
        assert csim.jt_isim_from_sum(c_total, c_objects) == 0


def test_jt_isim_from_sum_homogeneous() -> None:
    fps = np.zeros((100, 2048), dtype=np.uint8)
    c_total = fps.sum(0)
    c_objects = len(fps)
    assert pysim.jt_isim_from_sum(c_total, c_objects) == 1.0
    if CSIM_AVAIL:
        assert csim.jt_isim_from_sum(c_total, c_objects) == 1.0

    fps = np.ones((100, 2048), dtype=np.uint8)
    c_total = fps.sum(0)
    c_objects = len(fps)
    assert pysim.jt_isim_from_sum(c_total, c_objects) == 1.0
    if CSIM_AVAIL:
        assert csim.jt_isim_from_sum(c_total, c_objects) == 1.0


def test_jt_isim_from_sum_single() -> None:
    fps = make_fake_fingerprints(1, seed=17408390758220920002, pack=False)
    c_total = fps.sum(0)
    c_objects = len(fps)
    with pytest.warns(RuntimeWarning):
        _ = pysim.jt_isim_from_sum(c_total, c_objects)
    if CSIM_AVAIL:
        with pytest.warns(RuntimeWarning):
            _ = csim.jt_isim_from_sum(c_total, c_objects)


def test_jt_compl_isim() -> None:
    fps = make_fake_fingerprints(2, seed=17408390758220920002, pack=False)
    with pytest.warns(RuntimeWarning):
        _ = pysim.jt_compl_isim(fps)
    if CSIM_AVAIL:
        with pytest.warns(RuntimeWarning):
            _ = csim.jt_compl_isim(fps)

    fps = make_fake_fingerprints(10, seed=17408390758220920002, pack=False)
    output = pysim.jt_compl_isim(fps).tolist()
    assert output == snapshot(
        [
            0.20256457907452147,
            0.24748926949201983,
            0.22550084742079876,
            0.2002884861456855,
            0.23889840001690868,
            0.2364222674813306,
            0.1986207548061027,
            0.19904732709222533,
            0.21303348506016495,
            0.2225069540267648,
        ]
    )
    if CSIM_AVAIL:
        assert csim.jt_compl_isim(fps).tolist() == output
    assert (
        pysim.jt_compl_isim(np.zeros((10, 512), dtype=np.uint8))
        == np.ones(10, dtype=np.float64)
    ).all()
    if CSIM_AVAIL:
        assert (
            csim.jt_compl_isim(np.zeros((10, 512), dtype=np.uint8))
            == np.ones(10, dtype=np.float64)
        ).all()


def test_jt_isim_medoid() -> None:
    fps = make_fake_fingerprints(
        30, n_features=8, seed=17408390758220920002, pack=False
    )
    idx, m = bblean.similarity.jt_isim_medoid(fps)
    assert idx == snapshot(26)
    assert m.tolist() == snapshot([1, 1, 0, 1, 1, 1, 1, 1])



r"""Misc. utility functions"""

import os
from pathlib import Path
import itertools
import typing as tp
import sys
import subprocess
import platform
import importlib

import psutil
import numpy as np

__all__ = [
    "batched",
    "min_safe_uint",
    "cpp_extensions_are_enabled",
    "cpp_extensions_are_installed",
]

_T = tp.TypeVar("_T")


def min_safe_uint(nmax: int) -> np.dtype:
    r"""Returns the min uint dtype that holds a (positive) py int, excluding "object".

    Input must be a positive python integer.
    """
    out = np.min_scalar_type(nmax)
    # Check if the dtype is a pointer to a python bigint
    if out.hasobject:
        raise ValueError(f"n_samples: {nmax} is too large to hold in a uint64 array")
    return out


# Itertools recipe
def batched(iterable: tp.Iterable[_T], n: int) -> tp.Iterator[tuple[_T, ...]]:
    r"""Batch data into tuples of length n. The last batch may be shorter.

    This is equivalent to the batched receip from `itertools`.
    """
    # batched('ABCDEFG', 3) --> ('A', 'B', 'C') ('D', 'E', 'F') ('G',)
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(itertools.islice(it, n)):
        yield batch


def _import_bitbirch_variant(
    variant: str = "lean",
) -> tuple[tp.Any, tp.Callable[..., None]]:
    if variant not in ("lean", "int64", "uint8"):
        raise ValueError(f"Unknown variant {variant}")
    if variant == "lean":
        # Most up-to-date bb variant
        module = importlib.import_module("bblean.bitbirch")
    elif variant == "uint8":
        # Legacy variant of bb that uses uint8 and supports packing, but no extra optim
        module = importlib.import_module("bblean._legacy.bb_uint8")
    elif variant == "int64":
        # Legacy variant of bb that uses int64 fps (dense only)
        module = importlib.import_module("bblean._legacy.bb_int64")

    Cls = getattr(module, "BitBirch")
    fn = getattr(module, "set_merge")
    return Cls, fn


def _num_avail_cpus() -> int:
    if sys.platform == "darwin":
        # macOS doesn't expose cpu affinity, so assume all cpu's are available
        return os.cpu_count()
    return len(psutil.Process().cpu_affinity())


def _cpu_name() -> str:
    if sys.platform == "darwin":
        try:
            return subprocess.run(
                ["sysctl", "-n", "machdep.cpu.brand_string"],
                capture_output=True,
                text=True,
                check=True,
            ).stdout.strip()
        except Exception:
            pass

    if sys.platform == "linux":
        with open("/proc/cpuinfo") as f:
            for line in f:
                if line.startswith("model name"):
                    return line.split(":", 1)[1].strip()

    # Fallback for windows and all cases where it could not be found
    return platform.processor()


def _has_files_or_valid_symlinks(path: Path) -> bool:
    has_files = False
    for p in path.iterdir():
        if p.is_symlink() and not p.exists():
            return False

        if p.is_file():
            has_files = True
    return has_files


def cpp_extensions_are_enabled() -> bool:
    r"""Query whether the C++ BitBRICH extensions are currently enabled"""
    if os.getenv("BITBIRCH_NO_EXTENSIONS"):
        return False
    try:
        from bblean._cpp_similarity import jt_isim_from_sum  # noqa

        return True
    except ImportError:
        return False


def cpp_extensions_are_installed() -> bool:
    r"""Query whether the C++ BitBRICH extensions are currently installed"""
    try:
        from bblean._cpp_similarity import jt_isim_from_sum  # noqa

        return True
    except ImportError:
        return False
